@@ -108,6 +108,36 @@ if TYPE_CHECKING:
     from libqtile.core.manager import Qtile
 
 
+class ImplicitGrab(wlrq.HasListeners):
+    """Keep track of an implicit pointer grab.
+
+    A Wayland client expects to receive pointer events from the moment a
+    pointer button is pressed on its surface until the moment the button is
+    released.  The Wayland protocol leaves this behavior to the compositor.
+    """
+
+    def __init__(
+        self,
+        core: Core,
+        surface: Surface,
+        start_x: int,
+        start_y: int,
+        start_sx: int,
+        start_sy: int,
+    ) -> None:
+        self.core = core
+        self.surface = surface
+        self.start_dx = start_sx - start_x
+        self.start_dy = start_sy - start_y
+        self.add_listener(surface.destroy_event, self._on_destroy)
+
+    def finalize(self) -> None:
+        self.finalize_listeners()
+
+    def _on_destroy(self, _listener: Listener, _data: Any) -> None:
+        self.core._release_implicit_grab()
+
+
 class Core(base.Core, wlrq.HasListeners):
     supports_restarting: bool = False
 
@@ -198,6 +228,8 @@ class Core(base.Core, wlrq.HasListeners):
         self.cursor = Cursor(self.output_layout)
         self.cursor_manager = XCursorManager(24)
         self._gestures = PointerGesturesV1(self.display)
+        self._pressed_button_count = 0
+        self._implicit_grab: ImplicitGrab | None = None
         self.add_listener(self.seat.request_set_cursor_event, self._on_request_cursor)
         self.add_listener(self.cursor.axis_event, self._on_cursor_axis)
         self.add_listener(self.cursor.frame_event, self._on_cursor_frame)
@@ -339,16 +371,6 @@ class Core(base.Core, wlrq.HasListeners):
         # Start
         self.backend.start()
 
-        # Place cursor in middle of centre output
-        x = y = 0
-        box = self.output_layout.get_box()
-        if output := self.output_layout.output_at(box.width / 2, box.height / 2):
-            box = self.output_layout.get_box(reference=output, dest_box=box)
-            x = box.x + box.width / 2
-            y = box.y + box.height / 2
-        self.warp_pointer(x, y)
-        self.cursor_manager.set_cursor_image("left_ptr", self.cursor)
-
     @property
     def name(self) -> str:
         return "wayland"
@@ -403,6 +425,7 @@ class Core(base.Core, wlrq.HasListeners):
 
     def _on_start_drag(self, _listener: Listener, wlr_drag: Drag) -> None:
         logger.debug("Signal: seat start_drag")
+        self._release_implicit_grab()
         if not wlr_drag.icon:
             return
         self.live_dnd = wlrq.Dnd(self, wlr_drag)
@@ -436,9 +459,6 @@ class Core(base.Core, wlrq.HasListeners):
         output = Output(self, wlr_output)
         self.outputs.append(output)
 
-        if not self._current_output:
-            self._current_output = output
-
         # This is run during tests, when we want to fix the output's geometry
         if wlr_output.is_headless and "PYTEST_CURRENT_TEST" in os.environ:
             if len(self.outputs) == 1:
@@ -452,6 +472,17 @@ class Core(base.Core, wlrq.HasListeners):
         # Let the output layout place it
         self.output_layout.add_auto(wlr_output)
 
+        # Set the current output as we have none defined
+        # Now that we have our first output we can warp the pointer there too
+        # We also set the cursor image as we're initializing the cursor here anyways
+        if not self._current_output:
+            self._current_output = output
+            self.cursor_manager.set_cursor_image("left_ptr", self.cursor)
+            box = Box(*output.get_geometry())
+            x = box.x + box.width / 2
+            y = box.y + box.height / 2
+            self.warp_pointer(x, y)
+
     def _on_output_layout_change(self, _listener: Listener, _data: Any) -> None:
         logger.debug("Signal: output_layout change_event")
         config = OutputConfigurationV1()
@@ -543,7 +574,7 @@ class Core(base.Core, wlrq.HasListeners):
             box = xdg_surface.get_geometry()
             lx, ly = self.output_layout.closest_point(win.x + box.x, win.y + box.y)
             wlr_output = self.output_layout.output_at(lx, ly)
-            box = Box(*wlr_output.data.get_geometry())  # type: ignore[union-attr]
+            box = Box(*wlr_output.data.get_geometry())
             box.x = round(box.x - lx)
             box.y = round(box.y - ly)
             xdg_surface.popup.unconstrain_from_box(box)
@@ -551,10 +582,25 @@ class Core(base.Core, wlrq.HasListeners):
 
         logger.warning("xdg_shell surface had no role set. Ignoring.")
 
+    def _release_implicit_grab(self, time: int = 0) -> None:
+        if self._implicit_grab is not None:
+            logger.debug("Releasing implicit grab.")
+            self._implicit_grab.finalize()
+            self._implicit_grab = None
+            # Pretend the cursor just appeared where it is.
+            self._process_cursor_motion(time, self.cursor.x, self.cursor.y)
+
+    def _create_implicit_grab(self, time: int, surface: Surface, sx: float, sy: float) -> None:
+        self._release_implicit_grab(time)
+        logger.debug("Creating implicit grab.")
+        self._implicit_grab = ImplicitGrab(
+            self, surface, self.cursor.x, self.cursor.y, int(sx), int(sy)
+        )
+
     def _on_cursor_axis(self, _listener: Listener, event: pointer.PointerAxisEvent) -> None:
         handled = False
 
-        if event.delta != 0 and not self.exclusive_client:
+        if event.delta != 0 and not self.exclusive_client and not self._implicit_grab:
             # If we have a client who exclusively gets input, button bindings are disallowed.
             if event.orientation == pointer.AxisOrientation.VERTICAL:
                 button = 5 if 0 < event.delta else 4
@@ -577,9 +623,21 @@ class Core(base.Core, wlrq.HasListeners):
     def _on_cursor_button(self, _listener: Listener, event: pointer.PointerButtonEvent) -> None:
         assert self.qtile is not None
         self.idle.notify_activity(self.seat)
+        found = None
         pressed = event.button_state == input_device.ButtonState.PRESSED
         if pressed:
-            self._focus_by_click()
+            self._pressed_button_count += 1
+            if self._implicit_grab is None:
+                found = self._focus_by_click()
+        else:
+            if self._pressed_button_count > 0:  # sanity check
+                self._pressed_button_count -= 1
+
+        if self._implicit_grab is not None:
+            self.seat.pointer_notify_button(event.time_msec, event.button, event.button_state)
+            if self._pressed_button_count == 0:
+                self._release_implicit_grab(event.time_msec)
+            return
 
         handled = False
 
@@ -589,8 +647,18 @@ class Core(base.Core, wlrq.HasListeners):
             handled = self._process_cursor_button(button, pressed)
 
         if not handled:
+            if self._pressed_button_count == 1 and found and not self.live_dnd:
+                win, surface, sx, sy = found
+                if surface:
+                    self._create_implicit_grab(event.time_msec, surface, sx, sy)
             self.seat.pointer_notify_button(event.time_msec, event.button, event.button_state)
 
+    def _implicit_grab_motion(self, time: int) -> None:
+        if self._implicit_grab:
+            sx = self.cursor.x + self._implicit_grab.start_dx
+            sy = self.cursor.y + self._implicit_grab.start_dy
+            self.seat.pointer_notify_motion(time, sx, sy)
+
     def _on_cursor_motion(self, _listener: Listener, event: pointer.PointerMotionEvent) -> None:
         assert self.qtile is not None
         self.idle.notify_activity(self.seat)
@@ -616,7 +684,11 @@ class Core(base.Core, wlrq.HasListeners):
                 return
 
         self.cursor.move(dx, dy)
-        self._process_cursor_motion(event.time_msec, self.cursor.x, self.cursor.y)
+
+        if self._implicit_grab is None:
+            self._process_cursor_motion(event.time_msec, self.cursor.x, self.cursor.y)
+        else:
+            self._implicit_grab_motion(event.time_msec)
 
     def _on_cursor_motion_absolute(
         self, _listener: Listener, event: pointer.PointerMotionAbsoluteEvent
@@ -626,7 +698,10 @@ class Core(base.Core, wlrq.HasListeners):
 
         x, y = self.cursor.absolute_to_layout_coords(event.pointer.base, event.x, event.y)
         self.cursor.move(x - self.cursor.x, y - self.cursor.y)
-        self._process_cursor_motion(event.time_msec, self.cursor.x, self.cursor.y)
+        if self._implicit_grab is None:
+            self._process_cursor_motion(event.time_msec, self.cursor.x, self.cursor.y)
+        else:
+            self._implicit_grab_motion(event.time_msec)
 
     def _on_cursor_pinch_begin(
         self,
@@ -882,6 +957,11 @@ class Core(base.Core, wlrq.HasListeners):
                 self.output_layout.add(wlr_output, state.x, state.y)
                 wlr_output.set_transform(state.transform)
                 wlr_output.set_scale(state.scale)
+                # Ensure we have cursors loaded for the new scale factor.
+                self.cursor_manager.load(state.scale)
+                # Rescale the cursor if necessary
+                if not self.seat.pointer_state.focused_surface:
+                    self.cursor_manager.set_cursor_image("left_ptr", self.cursor)
             else:
                 if wlr_output.enabled:
                     wlr_output.enable(enable=False)
@@ -986,20 +1066,15 @@ class Core(base.Core, wlrq.HasListeners):
                     self.seat.pointer_notify_clear_focus()
 
             if win is not self.qtile.current_window:
-                if isinstance(win, window.Static):
-                    if self._hovered_window is not win:
-                        # qtile.current_window will never be a static window, but we
-                        # still only want to fire client_mouse_enter once, so check
-                        # self._hovered_window.
-                        hook.fire("client_mouse_enter", win)
-
-                    if motion is not None and self.qtile.config.follow_mouse_focus:
-                        self.qtile.focus_screen(win.screen.index, False)
-
-                else:
+                if self._hovered_window is not win:
+                    # We only want to fire client_mouse_enter once, so check
+                    # self._hovered_window.
                     hook.fire("client_mouse_enter", win)
 
-                    if motion is not None and self.qtile.config.follow_mouse_focus:
+                if motion is not None and self.qtile.config.follow_mouse_focus:
+                    if isinstance(win, window.Static):
+                        self.qtile.focus_screen(win.screen.index, False)
+                    else:
                         if win.group and win.group.current_window != win:
                             win.group.focus(win, False)
                         if (
@@ -1128,10 +1203,10 @@ class Core(base.Core, wlrq.HasListeners):
 
         assert self.qtile is not None
 
-        for win in self.qtile.windows_map.values():
-            if not isinstance(win, window.Window):
-                continue
-
+        managed_wins = [
+            w for w in self.qtile.windows_map.values() if isinstance(w, window.Window)
+        ]
+        for win in managed_wins:
             group = None
             if win.group:
                 if win.group.name in self.qtile.groups_map:
@@ -1253,12 +1328,12 @@ class Core(base.Core, wlrq.HasListeners):
             if keyboard := self.seat.keyboard:
                 self.seat.keyboard_notify_enter(surface, keyboard)
 
-    def _focus_by_click(self) -> None:
+    def _focus_by_click(self) -> tuple[window.WindowType, Surface | None, float, float] | None:
         assert self.qtile is not None
         found = self._under_pointer()
 
         if found:
-            win, _, _, _ = found
+            win, surface, sx, sy = found
 
             if self.exclusive_client:
                 # If we have a client who exclusively gets input, no other client's
@@ -1267,15 +1342,13 @@ class Core(base.Core, wlrq.HasListeners):
                     self.exclusive_client
                 ):
                     logger.debug("Focus withheld from window not owned by exclusive client.")
-                    return
+                    return None
 
             if self.qtile.config.bring_front_click is True:
                 win.bring_to_front()
             elif self.qtile.config.bring_front_click == "floating_only":
                 if isinstance(win, base.Window) and win.floating:
                     win.bring_to_front()
-                elif isinstance(win, base.Static):
-                    win.bring_to_front()
 
             if isinstance(win, window.Static):
                 if win.screen is not self.qtile.current_screen:
@@ -1291,6 +1364,8 @@ class Core(base.Core, wlrq.HasListeners):
             if screen:
                 self.qtile.focus_screen(screen.index, warp=False)
 
+        return found
+
     def _under_pointer(self) -> tuple[window.WindowType, Surface | None, float, float] | None:
         """
         Find which window and surface is currently under the pointer, if any.
